import logging
import os

import numpy as np

logger = logging.getLogger(__name__)

gpu_module = dict()


def compare(a, b):
    if type(a) != type(b):
        return False
    if type(a) is list or type(a) is tuple:
        if len(a) != len(b):
            return False
        comparisons = [compare(av, bv) for av, bv in zip(a, b)]
        if False in comparisons:
            return False
        return True
    if type(a) is np.ndarray:
        res = a == b
        if type(res) is np.ndarray:
            return res.all()
        else:
            return res
    return a == b


def get_existing_module(tmpl_name, tmpl_args):
    if tmpl_name not in gpu_module:
        return None

    existing_modules = gpu_module[tmpl_name]
    for module_info in existing_modules:
        tmpl_args_match = True
        for k, v in module_info["tmpl_args"].items():
            if not compare(v, tmpl_args[k]):
                tmpl_args_match = False
                break
        if tmpl_args_match:
            return module_info["module"]

    return None


def get_template(tmpl_name, tmpl_dir):
    import mako.lookup

    template_dirs = [os.getcwd()]
    if tmpl_dir is not None:
        template_dirs.append(tmpl_dir)
    lookup = mako.lookup.TemplateLookup(directories=template_dirs)
    return lookup.get_template(tmpl_name)


def template_with_mako(backend, preamble, tmpl, tmpl_args):
    try:
        return tmpl.render(**tmpl_args, backend=backend, preamble=preamble)
    except:  # noqa: E722
        # bare except is okay because we re-raise immediately
        import mako.exceptions

        logger.error(mako.exceptions.text_error_template().render())
        raise


def load(
    backend,
    preamble,
    compiler,
    tmpl_name,
    tmpl_dir=None,
    save_code=False,
    no_caching=False,
    tmpl_args=None,
):
    if tmpl_args is None:
        tmpl_args = dict()

    if not no_caching and not save_code:
        existing_module = get_existing_module(tmpl_name, tmpl_args)
        if existing_module is not None:
            logger.debug("returning cached gpu module " + tmpl_name)
            return existing_module

    tmpl = get_template(tmpl_name, tmpl_dir)
    if tmpl_args is None:
        tmpl_args = dict()

    code = template_with_mako(backend, preamble, tmpl, tmpl_args)
    rendered_fp = os.path.join(tmpl_dir, tmpl_name + ".rendered")
    with open(rendered_fp, "w") as f:
        f.write("\n")
        f.write(code)

    logger.debug("start compiling " + tmpl_name)

    module_info = dict()
    module_info["tmpl_args"] = tmpl_args
    module_info["module"] = compiler(code)

    gpu_module[tmpl_name] = gpu_module.get(tmpl_name, []) + [module_info]
    return module_info["module"]



from .aca import call_clu_aca
from .coordinators import (
    DISP_HS,
    STRAIN_HS,
    call_clu,
    call_clu_block,
    call_clu_free,
    call_clu_matrix,
)
from .geometry import strain_to_stress  # noqa: F401

DISP_SPEC = DISP_HS
STRAIN_SPEC = STRAIN_HS


def disp(obs_pts, tris, slips, nu):
    return call_clu(obs_pts, tris, slips, nu, DISP_HS)


def strain(obs_pts, tris, slips, nu):
    return call_clu(obs_pts, tris, slips, nu, STRAIN_HS)


def disp_matrix(obs_pts, tris, nu):
    return call_clu_matrix(obs_pts, tris, nu, DISP_HS)


def strain_matrix(obs_pts, tris, nu):
    return call_clu_matrix(obs_pts, tris, nu, STRAIN_HS)


def disp_free(obs_pts, tris, slips, nu):
    return call_clu_free(obs_pts, tris, slips, nu, DISP_HS)


def strain_free(obs_pts, tris, slips, nu):
    return call_clu_free(obs_pts, tris, slips, nu, STRAIN_HS)


def disp_block(obs_pts, tris, obs_start, obs_end, src_start, src_end, nu):
    return call_clu_block(
        obs_pts, tris, obs_start, obs_end, src_start, src_end, nu, DISP_HS
    )


def strain_block(obs_pts, tris, obs_start, obs_end, src_start, src_end, nu):
    return call_clu_block(
        obs_pts, tris, obs_start, obs_end, src_start, src_end, nu, STRAIN_HS
    )


def disp_aca(obs_pts, tris, obs_start, obs_end, src_start, src_end, nu, tol, max_iter):
    return call_clu_aca(
        obs_pts,
        tris,
        obs_start,
        obs_end,
        src_start,
        src_end,
        nu,
        tol,
        max_iter,
        DISP_HS,
    )


def strain_aca(
    obs_pts, tris, obs_start, obs_end, src_start, src_end, nu, tol, max_iter
):
    return call_clu_aca(
        obs_pts,
        tris,
        obs_start,
        obs_end,
        src_start,
        src_end,
        nu,
        tol,
        max_iter,
        STRAIN_HS,
    )



import logging

import pycuda
import pycuda.compiler
import pycuda.gpuarray

from .gpu_backend import load

logger = logging.getLogger(__name__)

cuda_initialized = False


def ensure_initialized():
    global cuda_initialized
    if not cuda_initialized:
        cuda_initialized = True
        import pycuda.autoinit

        gpu_idx = pycuda.driver.Context.get_device().get_attribute(
            pycuda._driver.device_attribute.MULTI_GPU_BOARD_GROUP_ID
        )
        logger.info("Initialized CUDA on gpu: " + str(gpu_idx))


def ptr(arr):
    if type(arr) is pycuda.gpuarray.GPUArray:
        return arr.gpudata
    return arr


def to(arr, float_type):
    ensure_initialized()
    if type(arr) is pycuda.gpuarray.GPUArray:
        return arr
    to_type = arr.astype(float_type)
    return pycuda.gpuarray.to_gpu(to_type)


def empty(shape, float_type):
    ensure_initialized()
    return pycuda.gpuarray.empty(shape, float_type)


def zeros(shape, float_type):
    ensure_initialized()
    return pycuda.gpuarray.zeros(shape, float_type)


def get(arr):
    return arr.get()


class CUDAContextWrapper(object):
    def __init__(self, context):
        self.ctx = context

    def __enter__(self):
        self.ctx.push()
        return self

    def __exit__(self, evalue, etype, etraceback):
        self.ctx.pop()


class ModuleWrapper:
    def __init__(self, module):
        self.module = module

    def __getattr__(self, name):
        kernel = self.module.get_function(name)

        def wrapper(*args):
            arg_ptrs = [ptr(a) for a in args[:-2]]
            grid = args[-2]
            block = args[-1]
            return kernel(*arg_ptrs, grid=grid, block=block)

        return wrapper


def compile(code):
    ensure_initialized()
    compiler_args = ["--use_fast_math", "--restrict"]
    return ModuleWrapper(pycuda.compiler.SourceModule(code, options=compiler_args))


def max_block_size(requested):
    return requested


def load_module(
    tmpl_name, tmpl_dir=None, save_code=False, no_caching=False, tmpl_args=None
):
    return load(
        "cuda", preamble, compile, tmpl_name, tmpl_dir, save_code, no_caching, tmpl_args
    )


preamble = """
#include <stdio.h>
#define CUDA
// taken from pycuda._cluda
#define WITHIN_KERNEL __device__
#define KERNEL extern "C" __global__
#define GLOBAL_MEM /* empty */
#define LOCAL_MEM __shared__
#define LOCAL_MEM_DYNAMIC extern __shared__
#define LOCAL_MEM_ARG /* empty */
#define CONSTANT __constant__
#define INLINE __forceinline__
#define SIZE_T unsigned int
#define VSIZE_T unsigned int
// used to align fields in structures
#define ALIGN(bytes) __align__(bytes)

//IF EVER NEEDED, THESE CAN BE CONVERTED TO DEFINES SO THAT THERE IS LESS PER
// KERNEL OVERHEAD.
WITHIN_KERNEL SIZE_T get_local_id(unsigned int dim)
{
    if(dim == 0) return threadIdx.x;
    if(dim == 1) return threadIdx.y;
    if(dim == 2) return threadIdx.z;
    return 0;
}
WITHIN_KERNEL SIZE_T get_group_id(unsigned int dim)
{
    if(dim == 0) return blockIdx.x;
    if(dim == 1) return blockIdx.y;
    if(dim == 2) return blockIdx.z;
    return 0;
}
WITHIN_KERNEL SIZE_T get_local_size(unsigned int dim)
{
    if(dim == 0) return blockDim.x;
    if(dim == 1) return blockDim.y;
    if(dim == 2) return blockDim.z;
    return 1;
}
WITHIN_KERNEL SIZE_T get_num_groups(unsigned int dim)
{
    if(dim == 0) return gridDim.x;
    if(dim == 1) return gridDim.y;
    if(dim == 2) return gridDim.z;
    return 1;
}
WITHIN_KERNEL SIZE_T get_global_size(unsigned int dim)
{
    return get_num_groups(dim) * get_local_size(dim);
}
WITHIN_KERNEL SIZE_T get_global_id(unsigned int dim)
{
    return get_local_id(dim) + get_group_id(dim) * get_local_size(dim);
}
"""



import numpy as np


def strain_to_stress(strain, mu, nu):
    """
    Compute stress given strain like:
        stress = 2 * mu * strain + lambda * Id(3) * trace(strain)

    Parameters
    ----------
    strain : {array-like}, shape (n_tensors, 6)
        The strain tensors ordered like (e_xx, e_yy, e_zz, e_xy, e_xz, e_yz)
    mu : float
        Shear modulus
    nu : float
        Poisson ratio

    Returns
    -------
    stress : np.ndarray, shape (n_tensors, 6)
        The stress tensors ordered like (s_xx, s_yy, s_zz, s_xy, s_xz, s_yz)
    """
    lam = 2 * mu * nu / (1 - 2 * nu)
    trace = np.sum(strain[:, :3], axis=1)
    stress = np.empty_like(strain)
    stress[:, :3] = 2 * mu * strain[:, :3] + lam * trace[:, np.newaxis]
    stress[:, 3:] = 2 * mu * strain[:, 3:]
    return stress


def compute_normal_vectors(tri_pts) -> np.ndarray:
    """
    Compute normal vectors for each triangle.

    Parameters
    ----------
    tri_pts : {array-like}, shape (n_triangles, 3, 3)
        The vertices of the triangles.

    Returns
    -------
    normals : np.ndarray, shape (n_triangles, 3)
        The normal vectors for each triangle.
    """
    leg1 = tri_pts[:, 1] - tri_pts[:, 0]
    leg2 = tri_pts[:, 2] - tri_pts[:, 0]
    # The normal vector is one axis of the TDCS and can be
    # computed as the cross product of the two corner-corner tangent vectors.
    Vnormal = np.cross(leg1, leg2, axis=1)
    # And it should be normalized to have unit length of course!
    Vnormal /= np.linalg.norm(Vnormal, axis=1)[:, None]
    return Vnormal


def compute_projection_transforms(origins, transformer) -> np.ndarray:
    """
    Convert vectors from one coordinate system to another. Unlike positions,
    this cannot be done with a simple pyproj call. We first need to set up a
    vector start and end point, convert those into the new coordinate system
    and then recompute the direction/distance between the start and end point.

    The output matrices are not pure rotation matrices because there is also
    a scaling of vector lengths. For example, converting from latitude to
    meters will result in a large scale factor.

    You can obtain the inverse transformation either by computing the inverse
    of the matrix or by passing an inverse pyproj.Transformer.

    Parameters
    ----------
    origins : {array-like}, shape (N, 3)
        The points at which we will compute rotation matrices
    transformer : pyproj.Transformer
        A pyproj.Transformer that will perform the necessary projection step.

    Returns
    -------
    transform_mats : np.ndarray, shape (n_triangles, 3, 3)
        The 3x3 rotation and scaling matrices that transform vectors from the
        EFCS to TDCS.
    """

    out = np.empty((origins.shape[0], 3, 3), dtype=origins.dtype)
    for d in range(3):
        eps = 1.0
        targets = origins.copy()
        targets[:, d] += eps
        proj_origins = np.array(
            transformer.transform(origins[:, 0], origins[:, 1], origins[:, 2])
        ).T.copy()
        proj_targets = np.array(
            transformer.transform(targets[:, 0], targets[:, 1], targets[:, 2])
        ).T.copy()
        out[:, :, d] = proj_targets - proj_origins
    return out


def compute_efcs_to_tdcs_rotations(tri_pts) -> np.ndarray:
    """
    Build rotation matrices that convert from an Earth-fixed coordinate system
    (EFCS) to a triangular dislocation coordinate system (TDCS).

    In the EFCS, the vectors will be directions/length in a map projection or
    an elliptical coordinate system.

    In the TDCS, the coordinates/vectors will be separated into:
    `(along-strike-distance, along-dip-distance, tensile-distance)`

    Note that in the Nikhoo and Walter 2015 and the Okada convention, the dip
    vector points upwards. This is different from the standard geologic
    convention where the dip vector points downwards.

    It may be useful to extract normal, dip or strike vectors from the rotation
    matrices that are returned by this function. The strike vectors are:
    `rot_mats[:, 0, :]`, the dip vectors are `rot_mats[:, 1, :]` and the normal
    vectors are `rot_mats[:, 2, :]`.

    To transform from TDCS back to EFCS, we simply need the transpose of the
    rotation matrices because the inverse of an orthogonal matrix is its
    transpose. To get this you can run `np.transpose(rot_mats, (0, 2, 1))`.

    Parameters
    ----------
    tri_pts : {array-like}, shape (n_triangles, 3, 3)
        The vertices of the triangles.

    Returns
    -------
    rot_mats : np.ndarray, shape (n_triangles, 3, 3)
        The 3x3 rotation matrices that transform vectors from the EFCS to TDCS.
    """
    Vnormal = compute_normal_vectors(tri_pts)
    eY = np.array([0, 1, 0])
    eZ = np.array([0, 0, 1])
    # The strike vector is defined as orthogonal to both the (0,0,1) vector and
    # the normal.
    Vstrike_raw = np.cross(eZ[None, :], Vnormal, axis=1)
    Vstrike_length = np.linalg.norm(Vstrike_raw, axis=1)

    # If eZ == Vnormal, we will get Vstrike = (0,0,0). In this case, just set
    # Vstrike equal to (0,±1,0).
    Vstrike = np.where(
        Vstrike_length[:, None] == 0, eY[None, :] * Vnormal[:, 2, None], Vstrike_raw
    )
    Vstrike /= np.linalg.norm(Vstrike, axis=1)[:, None]
    Vdip = np.cross(Vnormal, Vstrike, axis=1)
    return np.transpose(np.array([Vstrike, Vdip, Vnormal]), (1, 0, 2))



from .aca import call_clu_aca
from .coordinators import (
    DISP_FS,
    STRAIN_FS,
    call_clu,
    call_clu_block,
    call_clu_free,
    call_clu_matrix,
)
from .geometry import strain_to_stress  # noqa: F401
from .TDdispFS import TDdispFS

DISP_SPEC = DISP_FS
STRAIN_SPEC = STRAIN_FS


def py_disp(obs_pt, tri, slip, nu):
    return TDdispFS(obs_pt, tri, slip, nu)


def disp(obs_pts, tris, slips, nu):
    return call_clu(obs_pts, tris, slips, nu, DISP_FS)


def strain(obs_pts, tris, slips, nu):
    return call_clu(obs_pts, tris, slips, nu, STRAIN_FS)


def disp_matrix(obs_pts, tris, nu):
    return call_clu_matrix(obs_pts, tris, nu, DISP_FS)


def strain_matrix(obs_pts, tris, nu):
    return call_clu_matrix(obs_pts, tris, nu, STRAIN_FS)


def disp_free(obs_pts, tris, slips, nu):
    return call_clu_free(obs_pts, tris, slips, nu, DISP_FS)


def strain_free(obs_pts, tris, slips, nu):
    return call_clu_free(obs_pts, tris, slips, nu, STRAIN_FS)


def disp_block(obs_pts, tris, obs_start, obs_end, src_start, src_end, nu):
    return call_clu_block(
        obs_pts, tris, obs_start, obs_end, src_start, src_end, nu, DISP_FS
    )


def strain_block(obs_pts, tris, obs_start, obs_end, src_start, src_end, nu):
    return call_clu_block(
        obs_pts, tris, obs_start, obs_end, src_start, src_end, nu, STRAIN_FS
    )


def disp_aca(obs_pts, tris, obs_start, obs_end, src_start, src_end, nu, tol, max_iter):
    return call_clu_aca(
        obs_pts,
        tris,
        obs_start,
        obs_end,
        src_start,
        src_end,
        nu,
        tol,
        max_iter,
        DISP_FS,
    )


def strain_aca(
    obs_pts, tris, obs_start, obs_end, src_start, src_end, nu, tol, max_iter
):
    return call_clu_aca(
        obs_pts,
        tris,
        obs_start,
        obs_end,
        src_start,
        src_end,
        nu,
        tol,
        max_iter,
        STRAIN_FS,
    )



import logging
import os
import warnings

import pyopencl
import pyopencl.array

from .gpu_backend import load

logger = logging.getLogger(__name__)

gpu_initialized = False
gpu_ctx = None
gpu_queue = None


def report_devices(ctx):
    device_names = [d.name for d in ctx.devices]
    logger.info("initializing opencl context with devices = " + str(device_names))


def initialize_with_ctx(ctx):
    global gpu_initialized, gpu_ctx, gpu_queue
    gpu_ctx = ctx
    gpu_queue = pyopencl.CommandQueue(gpu_ctx)
    gpu_initialized = True

    report_devices(ctx)


def avoid_apple_cpu(ctx):
    """
    The Apple CPU OpenCL implementation is awful. Instead, we should just use
    PoCL.
    """
    if ctx.devices[0].platform.name == "Apple" and "CPU" in ctx.devices[0].name:
        platforms = pyopencl.get_platforms()
        platform_idx = None
        for i, p in enumerate(platforms):
            if p.name != "Apple":
                platform_idx = i
            else:
                apple_platform_idx = i
        if platform_idx is not None:
            warnings.warn(
                "The OpenCL context created used the Apple CPU"
                " implementation which is not supported. Trying again"
                f" with a different platform: {p.name}"
            )
            return pyopencl.create_some_context(answers=[str(platform_idx)])

        # If no other platforms were found, let's try to
        # find a non-CPU device like an Iris Pro.
        platform_idx = apple_platform_idx
        device_idx = None
        for i, d in enumerate(platforms[platform_idx].get_devices()):
            if "CPU" in d.name:
                continue
            device_idx = i
            break

        if device_idx is not None:
            warnings.warn(
                "The OpenCL context created used the Apple CPU"
                " implementation which is not supported. Trying again"
                f" with a different device: {d.name}"
            )
            return pyopencl.create_some_context(
                answers=[str(platform_idx), str(device_idx)]
            )

        raise NotImplementedError(
            "cutde does not support the Apple CPU OpenCL implementation and no other"
            " platform or device was found. Please consult the cutde README"
        )
    return ctx


def ensure_initialized():
    global gpu_initialized
    if not gpu_initialized:
        ctx = pyopencl.create_some_context()
        ctx = avoid_apple_cpu(ctx)
        initialize_with_ctx(ctx)


def ptr(arr):
    if type(arr) is pyopencl.array.Array:
        return arr.data
    return arr


def to(arr, float_type):
    ensure_initialized()
    if type(arr) is pyopencl.array.Array:
        return arr
    to_type = arr.astype(float_type)
    return pyopencl.array.to_device(gpu_queue, to_type)


def zeros(shape, float_type):
    ensure_initialized()
    return pyopencl.array.zeros(gpu_queue, shape, float_type)


def empty(shape, float_type):
    ensure_initialized()
    return pyopencl.array.empty(gpu_queue, shape, float_type)


def get(arr):
    return arr.get()


class ModuleWrapper:
    def __init__(self, module):
        self.module = module

    def __getattr__(self, name):
        kernel = getattr(self.module, name)

        def provide_queue_wrapper(*args):
            grid = args[-2]
            block = args[-1]
            global_size = [b * g for b, g in zip(grid, block)]
            arg_ptrs = [ptr(a) for a in args[:-2]]
            return kernel(gpu_queue, global_size, block, *arg_ptrs)

        return provide_queue_wrapper


def compile(code):
    ensure_initialized()

    compile_options = []

    if "CUTDE_OPENCL_NO_OPTS" in os.environ:
        logger.debug("compiling opencl without optimizations.")
        compile_options.append("-cl-opt-disable")

    return ModuleWrapper(pyopencl.Program(gpu_ctx, code).build(options=compile_options))


def max_block_size(requested):
    return requested


def load_module(
    tmpl_name, tmpl_dir=None, save_code=False, no_caching=False, tmpl_args=None
):
    return load(
        "opencl",
        preamble,
        compile,
        tmpl_name,
        tmpl_dir,
        save_code,
        no_caching,
        tmpl_args,
    )


preamble = """
// taken from pyopencl._cluda
// 'static' helps to avoid the "no previous prototype for function" warning
#if __OPENCL_VERSION__ >= 120
#define WITHIN_KERNEL static
#else
#define WITHIN_KERNEL
#endif
#define KERNEL __kernel
#define GLOBAL_MEM __global
#define LOCAL_MEM __local
#define LOCAL_MEM_DYNAMIC __local
#define LOCAL_MEM_ARG __local
#define CONSTANT __constant
// INLINE is already defined in Beignet driver
#ifndef INLINE
#define INLINE inline
#endif
#define SIZE_T size_t
#define VSIZE_T size_t
// used to align fields in structures
#define ALIGN(bytes) __attribute__ ((aligned(bytes)))
#if defined(cl_khr_fp64)
#pragma OPENCL EXTENSION cl_khr_fp64: enable
#elif defined(cl_amd_fp64)
#pragma OPENCL EXTENSION cl_amd_fp64: enable
#endif
"""



import logging

import numpy as np

import cutde.cpp_backend_double
import cutde.cpp_backend_float

logger = logging.getLogger(__name__)


def to(arr, float_type):
    return arr.ravel().astype(float_type)


def zeros(shape, float_type):
    return np.zeros(shape, dtype=float_type)


def empty(shape, float_type):
    return np.empty(shape, dtype=float_type)


def get(arr):
    return arr


def max_block_size(requested):
    return 1


def load_module(
    tmpl_name, tmpl_dir=None, save_code=False, no_caching=False, tmpl_args=None
):
    if tmpl_args["float_type"] == "float":
        return cutde.cpp_backend_float
    else:
        return cutde.cpp_backend_double



from math import ceil

import numpy as np

import cutde.backend as backend
from cutde.coordinators import (
    check_inputs,
    placeholder,
    process_block_inputs,
    solve_types,
    source_dir,
)


def check_tol_max_iter(obs_start, tol, max_iter, float_type):

    tol = np.array(tol)
    if tol.shape[0] != obs_start.shape[0]:
        raise ValueError("The length of tol must match obs_start.")

    tol = np.ascontiguousarray(tol, dtype=float_type)

    max_iter = np.array(max_iter)
    if max_iter.shape[0] != obs_start.shape[0]:
        raise ValueError("The length of max_iter must match obs_start.")

    max_iter = np.ascontiguousarray(max_iter, dtype=np.int32)
    return tol, max_iter


def call_clu_aca(
    obs_pts,
    tris,
    obs_start,
    obs_end,
    src_start,
    src_end,
    nu,
    tol,
    max_iter,
    fnc,
    Iref0=None,
    Jref0=None,
):
    fnc_name, vec_dim = fnc
    check_inputs(obs_pts, tris, placeholder)
    float_type, (obs_pts, tris, _) = solve_types(obs_pts, tris, placeholder)
    obs_start, obs_end, src_start, src_end = process_block_inputs(
        obs_start, obs_end, src_start, src_end
    )
    tol, max_iter = check_tol_max_iter(obs_start, tol, max_iter, float_type)

    default_chunk_size = 512
    team_size = backend.max_block_size(32)
    n_blocks = obs_end.shape[0]

    verbose = False
    gpu_config = dict(float_type=backend.np_to_c_type(float_type), verbose=verbose)
    module = backend.load_module("aca.cu", tmpl_args=gpu_config, tmpl_dir=source_dir)

    n_chunks = int(ceil(n_blocks / default_chunk_size))
    appxs = []
    for i in range(n_chunks):
        chunk_start = i * default_chunk_size
        chunk_size = min(n_blocks - chunk_start, default_chunk_size)
        chunk_end = chunk_start + chunk_size

        n_obs_per_block = (
            obs_end[chunk_start:chunk_end] - obs_start[chunk_start:chunk_end]
        )
        n_src_per_block = (
            src_end[chunk_start:chunk_end] - src_start[chunk_start:chunk_end]
        )
        n_rows = n_obs_per_block * vec_dim
        n_cols = n_src_per_block * 3
        block_sizes = n_rows * n_cols

        # Storage for the U, V output matrices. These will be in a packed format.
        gpu_buffer = backend.empty(block_sizes.sum(), float_type)

        # Storage for temporary rows and columns: RIref, RJref, RIstar, RJstar
        fworkspace_per_block = n_cols + n_rows + 3 * n_cols + vec_dim * n_rows
        fworkspace_ends = np.cumsum(fworkspace_per_block)
        fworkspace_starts = fworkspace_ends - fworkspace_per_block
        gpu_fworkspace = backend.empty(fworkspace_ends[-1], float_type)
        gpu_fworkspace_starts = backend.to(fworkspace_starts, np.int32)

        # uv_ptrs forms arrays that point to the start of each U/V vector pairs in
        # the main output buffer
        uv_ptrs_size = np.minimum(n_rows, n_cols)
        uv_ptrs_ends = np.cumsum(uv_ptrs_size)
        uv_ptrs_starts = uv_ptrs_ends - uv_ptrs_size
        gpu_uv_ptrs_starts = backend.to(uv_ptrs_starts, np.int32)
        gpu_uv_ptrs = backend.empty(uv_ptrs_ends[-1], np.int32)
        gpu_iworkspace = backend.empty(uv_ptrs_ends[-1], np.int32)

        # Output space for specifying the number of terms used for each
        # approximation.
        gpu_n_terms = backend.empty(chunk_size, np.int32)

        # Storage space for a pointer to the next empty portion of the output
        # buffer.
        gpu_next_ptr = backend.zeros(1, np.int32)

        # The index of the starting reference rows/cols.
        if Iref0 is None:
            Iref0_chunk = np.random.randint(0, n_rows, size=chunk_size, dtype=np.int32)
        else:
            Iref0_chunk = Iref0[chunk_start:chunk_end]
        if Jref0 is None:
            Jref0_chunk = np.random.randint(0, n_cols, size=chunk_size, dtype=np.int32)
        else:
            Jref0_chunk = Jref0[chunk_start:chunk_end]
        gpu_Iref0 = backend.to(Iref0_chunk, np.int32)
        gpu_Jref0 = backend.to(Jref0_chunk, np.int32)

        gpu_obs_pts = backend.to(obs_pts, float_type)
        gpu_tris = backend.to(tris, float_type)
        gpu_obs_start = backend.to(obs_start[chunk_start:chunk_end], np.int32)
        gpu_obs_end = backend.to(obs_end[chunk_start:chunk_end], np.int32)
        gpu_src_start = backend.to(src_start[chunk_start:chunk_end], np.int32)
        gpu_src_end = backend.to(src_end[chunk_start:chunk_end], np.int32)
        gpu_tol = backend.to(tol, float_type)
        gpu_max_iter = backend.to(max_iter, np.int32)

        if verbose:
            print(f"gpu_buffer.shape = {gpu_buffer.shape}")
            print(f"gpu_uv_ptrs.shape = {gpu_uv_ptrs.shape}")
            print(f"gpu_n_terms.shape = {gpu_n_terms.shape}")
            print(f"gpu_next_ptr.shape = {gpu_next_ptr.shape}")
            print(f"gpu_fworkspace.shape = {gpu_fworkspace.shape}")
            print(f"gpu_iworkspace.shape = {gpu_iworkspace.shape}")
            print(f"gpu_uv_ptrs_starts.shape = {gpu_uv_ptrs_starts.shape}")
            print(f"gpu_Iref0.shape = {gpu_Iref0.shape}")
            print(f"gpu_Jref0.shape = {gpu_Jref0.shape}")
            print(f"obs_pts.shape = {obs_pts.shape}")
            print(f"tris.shape = {tris.shape}")
            print(f"gpu_obs_start.shape = {gpu_obs_start.shape}")
            print(f"gpu_obs_end.shape = {gpu_obs_end.shape}")
            print(f"gpu_src_start.shape = {gpu_src_start.shape}")
            print(f"gpu_src_end.shape = {gpu_src_end.shape}")

        getattr(module, "aca_" + fnc_name)(
            gpu_buffer,
            gpu_uv_ptrs,
            gpu_n_terms,
            gpu_next_ptr,
            gpu_fworkspace,
            gpu_iworkspace,
            gpu_uv_ptrs_starts,
            gpu_fworkspace_starts,
            gpu_Iref0,
            gpu_Jref0,
            gpu_obs_pts,
            gpu_tris,
            gpu_obs_start,
            gpu_obs_end,
            gpu_src_start,
            gpu_src_end,
            gpu_tol,
            gpu_max_iter,
            float_type(nu),
            (chunk_size, 1, 1),
            (team_size, 1, 1),
        )

        # post-process the buffer to collect the U, V vectors
        buffer = backend.get(gpu_buffer)
        uv_ptrs = backend.get(gpu_uv_ptrs)
        n_terms = backend.get(gpu_n_terms)
        for i in range(chunk_size):
            us = []
            vs = []
            uv_ptr0 = uv_ptrs_starts[i]
            ptrs = uv_ptrs[uv_ptr0 + np.arange(n_terms[i])]
            us = buffer[ptrs[:, None] + np.arange(n_rows[i])[None, :]]
            vs = buffer[
                ptrs[:, None] + np.arange(n_rows[i], n_rows[i] + n_cols[i])[None, :]
            ]
            appxs.append((us.T, vs))
    return appxs
